#include <ardour/route.h>

#include "ardour_ui.h"
#include "automation_time_axis.h"
#include "automation_line.h"
#include "public_editor.h"
#include "canvas-simplerect.h"
#include "canvas-waveview.h"
#include "selection.h"
#include "ghostregion.h"
#include "rgb_macros.h"
#include "automation_selectable.h"
#include "point_selection.h"

#include "i18n.h"

using namespace ARDOUR;
using namespace Gtk;
using namespace Editing;

static const gchar * small_x_xpm[] = {
"11 11 2 1",
" 	c None",
".	c #FFFFFF",
"           ",
"           ",
"  .     .  ",
"   .   .   ",
"    . .    ",
"     .     ",
"    . .    ",
"   .   .   ",
"  .     .  ",
"           ",
"           "};

AutomationTimeAxisView::AutomationTimeAxisView (Session& s, Route& r, PublicEditor& e, Widget* p, std::string nom, std::string nomparent)

	: AxisView (s), 
	  TimeAxisView (s, e, p),
	  route (r),
	  _name (nom),
	  height_button (_("h")),
	  clear_button (_("clear")),
	  active_button (_("active"))
{
	base_rect = gtk_canvas_item_new (GTK_CANVAS_GROUP(canvas_display),
					 gtk_canvas_simplerect_get_type(),
					 "x1", 0.0,
					 "y1", 0.0,
					 "x2", 1000000.0,
					 "outline_color_rgba", RGBA_TO_UINT(40,40,40,255),
					 /* outline ends and bottom */
					 "outline_what", (guint32) (0x1|0x2|0x8),
					 "fill_color_rgba", RGBA_TO_UINT(162,162,208,105),
					 NULL);

	gtk_signal_connect (GTK_OBJECT(base_rect), "event",
			    (GtkSignalFunc) PublicEditor::canvas_automation_track_event,
			    this);

	hide_button.add (*(manage (new Pixmap (small_x_xpm))));

	height_button.set_name ("TrackSizeButton");
	active_button.set_name ("TrackVisualButton");
	clear_button.set_name ("TrackVisualButton");
	hide_button.set_name ("TrackRemoveButton");

	/* rearrange the name display */

	/* we never show these for automation tracks, so make
	   life easier and remove them.
	*/

	name_hbox.remove (name_entry);
	name_hbox.remove (expansion_box);

	/* move the name label over a bit */
	string shortpname = _name;
	bool shortened = false;
	
	if (_name.length()) {
		if (shortpname.length() > 18) {
			shortpname = shortpname.substr (0, 16);
			shortpname += "...";
			shortened = true;
		}
	}
	name_label.set_text (shortpname);
	
	
	name_label.set_alignment (1.0, 0.5);

	Label* plugname;

	if (nomparent.length()) {

		/* limit the plug name string */

		string pname = nomparent;

		if (pname.length() > 14) {
			pname = pname.substr (0, 11);
			pname += "...";
			shortened = true;
		}

 		plugname = manage (new Label (pname));
		plugname->set_name (X_("TrackPlugName"));
		plugname->set_alignment (1.0, 0.5);
		name_label.set_name (X_("TrackParameterName"));
		controls_table.remove (name_hbox);
		controls_table.attach (*plugname, 1, 6, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);
		controls_table.attach (name_hbox, 1, 6, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);
	}

	if (shortened) {
		string tipname = nomparent;
		if (!tipname.empty()) {
			tipname += ": ";
		}
		tipname += _name;
		ARDOUR_UI::instance()->tooltips().set_tip(controls_ebox, tipname);
	}
	
	/* add the buttons */
	controls_table.attach (hide_button, 0, 1, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);
	controls_table.attach (height_button, 0, 1, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);

	//controls_table.attach (hide_button, 6, 7, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);
	//controls_table.attach (height_button, 6, 7, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);

	controls_table.attach (active_button, 7, 9, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);
	controls_table.attach (clear_button, 7, 9, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND);

	controls_table.show_all ();

	/* add some padding to the LHS */

	//controls_lhs_pad.set_usize (17,-1);
	//controls_lhs_pad.show ();

	height_button.clicked.connect (slot (*this, &AutomationTimeAxisView::height_clicked));
	clear_button.clicked.connect (slot (*this, &AutomationTimeAxisView::clear_clicked));
	hide_button.clicked.connect (slot (*this, &AutomationTimeAxisView::hide_clicked));

	controls_base_selected_name = X_("AutomationTrackControlsBaseSelected");
	controls_base_unselected_name = X_("AutomationTrackControlsBase");
	controls_ebox.set_name (controls_base_unselected_name);
	expansion_box.set_name (controls_base_unselected_name);

	controls_frame.set_shadow_type (GTK_SHADOW_ETCHED_OUT);
	
	line = 0;
}

AutomationTimeAxisView::~AutomationTimeAxisView ()
{
	if (line) {
		delete line;
	}
}

void
AutomationTimeAxisView::height_clicked ()
{
	popup_size_menu (0);
}

void
AutomationTimeAxisView::clear_clicked ()
{
	line->clear ();
}

void
AutomationTimeAxisView::set_height (TrackHeight h)
{
	bool changed = (height != (unsigned int) h);

	controls_table.show_all ();

	TimeAxisView::set_height (h);
	gtk_object_set (GTK_OBJECT(base_rect), "y2", (double) h, NULL);
	line->set_height (h);

	for (list<GhostRegion*>::iterator i = ghosts.begin(); i != ghosts.end(); ++i) {
		(*i)->set_height ();
	}

	name_entry.hide ();
	expansion_box.hide ();

	if (changed) {
		/* only emit the signal if the height really changed */
		 route.gui_changed ("track_height", (void *) 0); /* EMIT_SIGNAL */
	}
}

void
AutomationTimeAxisView::set_samples_per_unit (double spu)
{
	TimeAxisView::set_samples_per_unit (editor.get_current_zoom());
	line->reset ();
}
 
void
AutomationTimeAxisView::hide_clicked ()
{
	set_marked_for_display (false);
	hide ();
}


void
AutomationTimeAxisView::build_display_menu ()
{
	using namespace Menu_Helpers;

	/* get the size menu ready */

	build_size_menu ();

	/* prepare it */

	TimeAxisView::build_display_menu ();

	/* now fill it with our stuff */

	MenuList& items = display_menu->items();

	items.push_back (MenuElem (_("Height"), *size_menu));
	items.push_back (SeparatorElem());
	items.push_back (MenuElem (_("Hide"), slot (*this, &AutomationTimeAxisView::hide_clicked)));
}

bool
AutomationTimeAxisView::cut_copy_clear (Selection& selection, CutCopyOp op)
{
	AutomationList* what_we_got = 0;
	AutomationList& alist (line->the_list());
	bool ret = false;

	_session.add_undo (alist.get_memento());

	switch (op) {
	case Cut:
		if ((what_we_got = alist.cut (selection.time.front().start, selection.time.front().end)) != 0) {
			editor.get_cut_buffer().add (what_we_got);
			_session.add_redo_no_execute (alist.get_memento());
			ret = true;
		}
		break;
	case Copy:
		if ((what_we_got = alist.copy (selection.time.front().start, selection.time.front().end)) != 0) {
			editor.get_cut_buffer().add (what_we_got);
		}
		break;

	case Clear:
		if ((what_we_got = alist.cut (selection.time.front().start, selection.time.front().end)) != 0) {
			_session.add_redo_no_execute (alist.get_memento());
			delete what_we_got;
			what_we_got = 0;
			ret = true;
		}
		break;
	}

	if (what_we_got) {
		for (AutomationList::iterator x = what_we_got->begin(); x != what_we_got->end(); ++x) {
			double foo = (*x)->value;
			line->model_to_view_y (foo);
			(*x)->value = foo;
		}
	}

	return ret;
}

bool
AutomationTimeAxisView::cut_copy_clear_objects (PointSelection& selection, CutCopyOp op)
{
	AutomationList* what_we_got = 0;
	AutomationList& alist (line->the_list());
	bool ret = false;

	_session.add_undo (alist.get_memento());

	for (PointSelection::iterator i = selection.begin(); i != selection.end(); ++i) {

		if (&(*i).track != this) {
			continue;
		}

		switch (op) {
		case Cut:
			if ((what_we_got = alist.cut ((*i).start, (*i).end)) != 0) {
				editor.get_cut_buffer().add (what_we_got);
				_session.add_redo_no_execute (alist.get_memento());
				ret = true;
			}
			break;
		case Copy:
			if ((what_we_got = alist.copy ((*i).start, (*i).end)) != 0) {
				editor.get_cut_buffer().add (what_we_got);
			}
			break;
			
		case Clear:
			if ((what_we_got = alist.cut ((*i).start, (*i).end)) != 0) {
				_session.add_redo_no_execute (alist.get_memento());
				delete what_we_got;
				what_we_got = 0;
				ret = true;
			}
			break;
		}
	}
		
	if (what_we_got) {
		for (AutomationList::iterator x = what_we_got->begin(); x != what_we_got->end(); ++x) {
			double foo = (*x)->value;
			line->model_to_view_y (foo);
			(*x)->value = foo;
		}
	}

	return ret;
}

bool
AutomationTimeAxisView::paste (jack_nframes_t pos, float times, Selection& selection, size_t nth)
{
	AutomationList& alist (line->the_list());
	AutomationSelection::iterator p;

	for (p = selection.lines.begin(); p != selection.lines.end() && nth; ++p, --nth);

	if (p == selection.lines.end()) {
		return false;
	}

	/* Make a copy of the list because we have to scale the
	   values from view coordinates to model coordinates, and we're
	   not supposed to modify the points in the selection.
	*/
	   
	AutomationList copy (**p);

	for (AutomationList::iterator x = copy.begin(); x != copy.end(); ++x) {
		double foo = (*x)->value;
		line->view_to_model_y (foo);
		(*x)->value = foo;
	}

	_session.add_undo (alist.get_memento());
	alist.paste (copy, pos, times);
	_session.add_redo_no_execute (alist.get_memento());

	return true;
}

void
AutomationTimeAxisView::add_ghost (GhostRegion* gr)
{
	ghosts.push_back (gr);
}

void
AutomationTimeAxisView::remove_ghost (GhostRegion* gr)
{
	list<GhostRegion*>::iterator i;

	for (i = ghosts.begin(); i != ghosts.end(); ++i) {
		if ((*i) == gr) {
			ghosts.erase (i);
			break;
		}
	}
}

void
AutomationTimeAxisView::get_selectables (jack_nframes_t start, jack_nframes_t end, double top, double bot, list<Selectable*>& results)
{
	if (line && touched (top, bot)) {
		double topfrac;
		double botfrac;

		/* remember: this is X Window - coordinate space starts in upper left and moves down.
		   y_position is the "origin" or "top" of the track.
		*/

		double mybot = y_position + height; // XXX need to include Editor::track_spacing; 

		if (y_position >= top && mybot <= bot) {

			/* y_position is below top, mybot is above bot, so we're fully
			   covered vertically.
			*/

			topfrac = 1.0;
			botfrac = 0.0;

		} else {

			/* top and bot are within y_position .. mybot */

			topfrac = 1.0 - ((top - y_position) / height);
			botfrac = 1.0 - ((bot - y_position) / height);
		}

		line->get_selectables (start, end, botfrac, topfrac, results);
	}
}

void
AutomationTimeAxisView::set_selected_points (PointSelection& points)
{
	if (line) {
		line->set_selected_points (points);
	}
}
