/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: plugin_ui.cc,v 1.47 2004/02/16 04:26:18 pauld Exp $
*/

#include <climits>
#include <cerrno>
#include <cmath>
#include <string>

#include <pbd/stl_delete.h>
#include <pbd/xml++.h>

#include <gtkmmext/click_box.h>
#include <gtkmmext/fastmeter.h>
#include <gtkmmext/slider_controller.h>
#include <gtkmmext/utils.h>

#include <midi++/manager.h>

#include <gtk--/box.h>

#include <ardour/audioengine.h>
#include <ardour/ladspa_plugin.h>
#include <ardour/insert.h>

#include "ardour_ui.h"
#include "prompter.h"
#include "plugin_ui.h"
#include "doi.h"
#include "utils.h"

#include "i18n.h"

using namespace ARDOUR;
using namespace Gtkmmext;
using namespace Gtk;
using namespace SigC;

PluginUIWindow::PluginUIWindow (AudioEngine &engine, PluginInsert& insert, bool scrollable)
	: _pluginui (engine, insert, scrollable)
{
	set_position (GTK_WIN_POS_MOUSE);
	set_name ("PluginEditor");
	add_events (GDK_KEY_PRESS_MASK|GDK_KEY_RELEASE_MASK);

	add (_pluginui);

	delete_event.connect (bind (slot (just_hide_it), reinterpret_cast<Window*> (this)));
	insert.GoingAway.connect (slot (*this, &PluginUIWindow::plugin_going_away));
	map_event.connect (slot (_pluginui, &PluginUI::start_updating));
	unmap_event.connect (slot (_pluginui, &PluginUI::stop_updating));

	if ( scrollable ) {
		gint h = _pluginui.get_preferred_height ();
		if (h > 600) h = 600;
		set_default_size (450, h); 
	}
}

PluginUIWindow::~PluginUIWindow ()

{
}

PluginUI::PluginUI (AudioEngine &engine, PluginInsert& pi, bool scrollable)
	:engine(engine),
	 insert (pi),
	 plugin (insert.plugin()),
	 button_table (initial_button_rows, initial_button_cols),
	 output_table (initial_output_rows, initial_output_cols),
	 save_btn(_("Save")), load_btn(_("Load")),
	 bypass_button (_("bypass")),
	 is_scrollable(scrollable)
{
	set_name ("PluginEditor");
	set_border_width (10);
	set_homogeneous (false);

	combo.set_value_in_list(true,true);
	combo.set_use_arrows(true);
	combo.set_use_arrows_always(true);
	combo.set_popdown_strings(plugin.get_ladspa_settings());

	combo.get_entry()->focus_in_event.connect (slot (*this, &PluginUI::entry_focus_event));
	combo.get_entry()->focus_out_event.connect (slot (*this, &PluginUI::entry_focus_event));
	combo.get_entry()->activate.connect(slot (*this, &PluginUI::setting_selected));

	save_btn.set_name ("PluginSaveButton");
	save_btn.clicked.connect(slot (*this, &PluginUI::save_plugin_setting));

	load_btn.set_name ("PluginLoadButton");
	load_btn.clicked.connect(slot (*this, &PluginUI::setting_selected));
	
	bypass_button.set_name ("PluginBypassButton");
	bypass_button.toggled.connect (slot (*this, &PluginUI::bypass_toggled));

	settings_box.set_border_width (10);
	settings_box.set_homogeneous (false);

	HBox* constraint_hbox = manage (new HBox);
	VBox* constraint_vbox = manage (new VBox);
	HBox* smaller_hbox = manage (new HBox);

	smaller_hbox->pack_start (combo, false, false);
	smaller_hbox->pack_start (load_btn, false, false);
	smaller_hbox->pack_start (save_btn, false, false);

	constraint_hbox->set_spacing (10);
	constraint_hbox->pack_end (bypass_button, false, false);

	constraint_vbox->set_spacing (10);
	constraint_vbox->pack_start (*constraint_hbox, false, false);
	constraint_vbox->pack_start (*smaller_hbox, true, false);

	VBox* info_vbox = manage (new VBox);
	info_vbox->pack_start (nameinfo_label, false, false);
	info_vbox->pack_start (makerinfo_label, false, false);
	info_vbox->pack_start (paraminfo_label, false, false, 5);

	settings_box.pack_start (*info_vbox, false, false, 10);
	settings_box.pack_end (*constraint_vbox, false, false);

	pack_start (settings_box, false, false);

	if ( is_scrollable ) {
		scroller.set_policy (GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		scroller.set_name ("PluginEditor");
		scroller_view.set_name("PluginEditor");
		scroller_view.add (hpacker);
		scroller.add (scroller_view);
		
		pack_start (scroller, true, true);
	}
	else {
		pack_start (hpacker, false, false);
	}

	insert.active_changed.connect (slot (*this, &PluginUI::redirect_active_changed));
	bypass_button.set_active (!insert.active());
	
	build (engine);
}

PluginUI::~PluginUI ()
{
	if (output_controls.size() > 0) {
		screen_update_connection.disconnect();
	}
}

void
PluginUI::build (AudioEngine &engine)

{
	guint32 i = 0;
	guint32 x = 0;
	Frame* frame;
	Frame* bt_frame;
	VBox* box;
	char info[32];
	int output_row, output_col;
	int button_row, button_col;
	int output_rows, output_cols;
	int button_rows, button_cols;
	guint32 n_ins=0, n_outs = 0;

	prefheight = 30;
	hpacker.set_spacing (10);

	output_rows = initial_output_rows;
	output_cols = initial_output_cols;
	button_rows = initial_button_rows;
	button_cols = initial_button_cols;
	output_row = 0;
	button_row = 0;
	output_col = 0;
	button_col = 0;

	button_table.set_homogeneous (true);
	button_table.set_row_spacings (2);
	button_table.set_col_spacings (2);
	output_table.set_homogeneous (true);
	output_table.set_row_spacings (2);
	output_table.set_col_spacings (2);
	button_table.set_border_width (5);
	output_table.set_border_width (5);

	hpacker.set_border_width (10);

	bt_frame = manage (new Frame);
	bt_frame->set_name ("BaseFrame");
	bt_frame->add (button_table);
	hpacker.pack_start(*bt_frame, true, true);

	box = manage (new VBox);
	frame = manage (new Frame);
	frame->set_name ("BaseFrame");
	frame->add (*box);
	hpacker.pack_start(*frame, true, true);

	box->set_border_width (5);
	box->set_spacing (1);
	
	/* find all ports. build control elements for all appropriate control ports */

	for (i = 0; i < plugin.port_count(); ++i) {

		if (LADSPA_IS_PORT_CONTROL(plugin.port_descriptors()[i]))
		{
			ControlUI* cui;
				
			/* if we are scrollable, just use one long column */

			if (!is_scrollable) {
				if (x++ > 7){
					frame = manage (new Frame);
					frame->set_name ("BaseFrame");
					box = manage (new VBox);
					
					box->set_border_width (5);
					box->set_spacing (1);

					frame->add (*box);
					hpacker.pack_start(*frame,true,true);

					x = 0;
				}
			}

			/* Don't show latency control ports */

			if (!strcmp(plugin.port_names()[i], X_("latency"))) {
				continue;
			}
				
			if ((cui = build_control_ui (engine, i, plugin.get_nth_midi_control (i))) == 0) {
				error << compose(_("Plugin Editor: could not build control element for port %1"), i) << endmsg;
				continue;
			}
				
			if (cui->control || cui->clickbox) {

				box->pack_start (*cui, false, false);

			} else if (cui->button) {

				if (button_row == button_rows) {
					button_row = 0;
					if (++button_col == button_cols) {
						button_cols += 2;
						button_table.resize (button_rows, button_cols);
					}
				}

				button_table.attach (*cui, button_col, button_col + 1, button_row, button_row+1, 
						     GTK_FILL|GTK_EXPAND, 0);
				button_row++;

			} else if (cui->display) {

				if (output_row == output_rows) {
					output_row = 0;
					if (++output_col == output_cols) {
						output_cols += 2;
						output_table.resize (output_rows, output_cols);
					}
				}
				
				output_table.attach (*cui, output_col, output_col + 1, output_row, output_row+1, 
						     GTK_FILL|GTK_EXPAND, 0);
 
				output_row++;
			}
				
			/* HACK: ideally the preferred height would be queried from
			   the complete hpacker, but I can't seem to get that
			   information in time, so this is an estimation 
			*/

			prefheight += 30;

		} else if ( LADSPA_IS_PORT_AUDIO (plugin.port_descriptors()[i]) ) {

			if ( LADSPA_IS_PORT_INPUT (plugin.port_descriptors()[i]) ) {
				n_ins++;
			}  else if ( LADSPA_IS_PORT_OUTPUT (plugin.port_descriptors()[i]) ) {
				n_outs++;
			}
		}
		
	}

	if (box->children().empty()) {
		hpacker.remove (*frame);
	}

	if (button_table.children().empty()) {
		hpacker.remove (*bt_frame);
	}

	if (!output_table.children().empty()) {
		frame = manage (new Frame);
		frame->set_name ("BaseFrame");
		frame->add (output_table);
		hpacker.pack_end (*frame, true, true);
	}

	output_update ();

	nameinfo_label.set_text (plugin.name());
	nameinfo_label.set_name ("PluginNameInfo");
	nameinfo_label.set_alignment (0.0, 0.0);

	string maker = plugin.maker();
	string::size_type email_pos;

	if ((email_pos = maker.find_first_of ('<')) != string::npos) {
		maker = maker.substr (0, email_pos - 1);
	}

	if (maker.length() > 32) {
		maker = maker.substr (0, 32);
		maker += " ...";
	}

	makerinfo_label.set_text (maker);
	makerinfo_label.set_name ("PluginMakerInfo");
	makerinfo_label.set_alignment (0.0, 0.0);
	
	snprintf (info, sizeof(info),"[ %u %s | %u %s ]",
		  n_ins, ngettext(_("in"), _("ins"), n_ins),
		  n_outs, ngettext(_("out"), _("outs"), n_outs));
	paraminfo_label.set_text (info);
	paraminfo_label.set_name ("PluginParameterInfo");
	paraminfo_label.set_alignment (0.0, 0.0);

	output_table.show_all ();
	button_table.show_all ();
}

PluginUI::ControlUI::ControlUI ()
	: automate_rec_button (_("arec")),
	  automate_play_button (_("aplay"))
{
	automate_rec_button.set_name ("PluginAutomateRecordButton");
	automate_play_button.set_name ("PluginAutomatePlayButton");
	ignore_change = 0;
	display = 0;
	button = 0;
	control = 0;
	clickbox = 0;
	adjustment = 0;
	meterinfo = 0;
}

PluginUI::ControlUI::~ControlUI() 
{
	if (adjustment) {
		delete adjustment;
	}

	if (meterinfo) {
		delete meterinfo->meter;
		delete meterinfo;
	}
}
static void integer_printer (char buf[32], Adjustment &adj, void *arg)
{
	snprintf (buf, 32, "%.0f", adj.get_value());
}

PluginUI::ControlUI*
PluginUI::build_control_ui (AudioEngine &engine, guint32 port_index, MIDI::Controllable* mcontrol)

{
	LADSPA_PortDescriptor desc;
	ControlUI* control_ui;
	LADSPA_PortRangeHint prh;

	desc = plugin.port_descriptors()[port_index];
	prh  = plugin.port_range_hints()[port_index];

	control_ui = manage (new ControlUI ());
	control_ui->port_index = port_index;
	control_ui->update_pending = false;
	control_ui->label.set_text (plugin.port_names()[port_index]);
	control_ui->label.set_alignment (0.0, 0.5);
	control_ui->label.set_name ("PluginParameterLabel");

	control_ui->set_spacing (5);

	if (LADSPA_IS_PORT_INPUT(desc)) {

		if (LADSPA_IS_HINT_TOGGLED(prh.HintDescriptor)) {

			/* Build a button */
		
			control_ui->adjustment = 0;
			control_ui->button = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->toggle_button_pix(),
									 "PluginEditorButton", false);

			/* XXX memory leak: button not destroyed by ControlUI
			   destructor, and manage() reports object hierarchy
			   ambiguity.
			*/

			control_ui->pack_start (control_ui->label, true, true);
			control_ui->pack_start (*control_ui->button, false, true);
			control_ui->pack_start (control_ui->automate_rec_button, false, false);
			control_ui->pack_start (control_ui->automate_play_button, false, false);

			control_ui->button->clicked.connect (bind (slot (*this, &PluginUI::control_port_toggled), control_ui));
		
			if(plugin.get_control_port(port_index) == 1){
				control_ui->button->set_active(true);
			}

			return control_ui;
		}
	
		control_ui->adjustment = new Adjustment (0, 0, 0, 0, 0);

		/* XXX this code is not right yet, because it doesn't handle
		   the absence of bounds in any sensible fashion.
		*/
		
		if (LADSPA_IS_HINT_BOUNDED_BELOW(prh.HintDescriptor)) {
			if (LADSPA_IS_HINT_SAMPLE_RATE(prh.HintDescriptor)) {
				control_ui->adjustment->set_lower (prh.LowerBound * engine.frame_rate());
			} else {
				control_ui->adjustment->set_lower (prh.LowerBound);
			}
		} else {
			control_ui->adjustment->set_lower (0);
		}

		if (LADSPA_IS_HINT_BOUNDED_ABOVE(prh.HintDescriptor)) {
			if (LADSPA_IS_HINT_SAMPLE_RATE(prh.HintDescriptor)) {
				control_ui->adjustment->set_upper (prh.UpperBound * engine.frame_rate());
			} else {
				control_ui->adjustment->set_upper (prh.UpperBound);
			}
		} else {
			control_ui->adjustment->set_upper (4);
		}

		control_ui->logarithmic = LADSPA_IS_HINT_LOGARITHMIC(prh.HintDescriptor);

		if (control_ui->logarithmic) {
			if (control_ui->adjustment->get_lower() == 0.0) {
				control_ui->adjustment->set_lower (control_ui->adjustment->get_upper()/10000);
			}
			control_ui->adjustment->set_upper (log(control_ui->adjustment->get_upper()));
			control_ui->adjustment->set_lower (log(control_ui->adjustment->get_lower()));
		}
	
		float delta = control_ui->adjustment->get_upper() - control_ui->adjustment->get_lower();

		control_ui->adjustment->set_page_size (delta/100.0);

		if (LADSPA_IS_HINT_INTEGER(prh.HintDescriptor)) {
			control_ui->adjustment->set_step_increment (1.0);
			control_ui->adjustment->set_page_increment (10.0);

			control_ui->clickbox = new ClickBox (control_ui->adjustment, "PluginUIClickBox");
			Gtkmmext::set_usize_to_display_given_text (*control_ui->clickbox, "g9999999", 2, 2);
			control_ui->clickbox->set_print_func (integer_printer, 0);
		} else {
			control_ui->adjustment->set_step_increment (delta/1000.0);
			control_ui->adjustment->set_page_increment (delta/10.0);
			
			control_ui->control = new HSliderController (ARDOUR_UI::instance()->plugin_ui_slider_pix(),
								     control_ui->adjustment,
								     mcontrol);

			control_ui->control->get_spin_button().focus_in_event.connect (slot (*this, &PluginUI::entry_focus_event));
			control_ui->control->get_spin_button().focus_out_event.connect (slot (*this, &PluginUI::entry_focus_event));

		}

		if (control_ui->logarithmic) {
			control_ui->adjustment->set_value(log(plugin.get_control_port(port_index)));
		} else{
			control_ui->adjustment->set_value(plugin.get_control_port(port_index));
		}

		/* XXX memory leak: SliderController not destroyed by ControlUI
		   destructor, and manage() reports object hierarchy
		   ambiguity.
		*/

		control_ui->pack_start (control_ui->label, true, true);
		if (LADSPA_IS_HINT_INTEGER(prh.HintDescriptor)) {
			control_ui->pack_start (*control_ui->clickbox, false, false);
		} else {
			control_ui->pack_start (*control_ui->control, false, false);
		}
		control_ui->pack_start (control_ui->automate_rec_button, false, false);
		control_ui->pack_start (control_ui->automate_play_button, false, false);

		control_ui->automate_play_button.set_active(insert.get_port_automation_playback (port_index));
		control_ui->automate_rec_button.set_active(insert.get_port_automation_record (port_index));
		
		control_ui->adjustment->value_changed.connect (bind (slot (*this, &PluginUI::control_adjustment_changed), control_ui));
		control_ui->automate_rec_button.toggled.connect (bind (slot (*this, &PluginUI::arec_toggled), control_ui, (unsigned long) port_index));
		control_ui->automate_play_button.toggled.connect (bind (slot (*this, &PluginUI::aplay_toggled), control_ui, (unsigned long) port_index));

	} else if (LADSPA_IS_PORT_OUTPUT(desc)) {
		
		control_ui->display = manage (new EventBox);
		control_ui->display->set_name ("ParameterValueDisplay");

		control_ui->display_label = manage (new Label);
		control_ui->display_label->set_name ("ParameterValueDisplay");

		control_ui->display->add (*control_ui->display_label);
		Gtkmmext::set_usize_to_display_given_text (*control_ui->display, "g999999", 2, 2);

		control_ui->display->show_all ();

		/* set up a meter */
		/* TODO: only make a meter if the port is Hinted for it */

		MeterInfo * info = new MeterInfo(port_index);
 		control_ui->meterinfo = info;
		
		info->meter = new FastMeter (100, 5, FastMeter::Horizontal);

 		if (LADSPA_IS_HINT_BOUNDED_BELOW(prh.HintDescriptor)) {
 			info->min = prh.LowerBound;
 		} else {
			info->min_unbound = true;
		}
		
 		if (LADSPA_IS_HINT_BOUNDED_ABOVE(prh.HintDescriptor)) {
 			info->max = prh.UpperBound;
 		} else {
			info->max_unbound = true;
		}

		control_ui->vbox = manage (new VBox);

 		control_ui->vbox->pack_start (control_ui->label, true, true);
 		control_ui->vbox->pack_start (*info->meter, true, true);
		
 		control_ui->pack_start (*control_ui->vbox, true, true);
 		control_ui->pack_start (*control_ui->display, false, false);

		control_ui->meterinfo->meter->show_all();
		control_ui->meterinfo->packed = true;
		
		output_controls.push_back (control_ui);
	}
	
	plugin.ControlPortChanged.connect (bind (slot (*this, &PluginUI::control_port_changed), control_ui));
	return control_ui;
}

void
PluginUI::arec_toggled (ControlUI* cui, unsigned long port)
{
	bool x;

	if ((x = cui->automate_rec_button.get_active()) != insert.get_port_automation_record (port)) {
		if (x) {
			insert.set_port_automation_record (port, AutomationList::Write);
		} else {
			insert.set_port_automation_record (port, AutomationList::Off);
		}
	}
}

void
PluginUI::aplay_toggled (ControlUI* cui, unsigned long port)
{
	bool x;

	if ((x = cui->automate_play_button.get_active()) != insert.get_port_automation_playback (port)) {
		insert.set_port_automation_playback (port, x);
	}
}

void
PluginUI::control_adjustment_changed (ControlUI* cui)
{
	if (cui->ignore_change) {
		return;
	}

	double value = cui->adjustment->get_value();

	if (cui->logarithmic) {
	  	value = exp(value);
	}

	insert.set_control_port (cui->port_index, (LADSPA_Data) value);
}

void
PluginUI::control_port_changed (unsigned long abs_port_id, LADSPA_Data val, ControlUI* cui)
{
	if (cui->port_index == abs_port_id) {
		if (!cui->update_pending) {
			cui->update_pending = true;
			Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &PluginUI::update_control_display), cui));
		}
	}
}

void
PluginUI::update_control_display (ControlUI* cui)	
{
	/* XXX how do we handle logarithmic stuff here ? */
	
	cui->update_pending = false;

	LADSPA_Data val = plugin.get_control_port (cui->port_index);

	cui->ignore_change++;
	if (cui->adjustment == 0) {

		if (val > 0.5) {
			cui->button->set_active (true);
		} else {
			cui->button->set_active (false);
		}

	} else {
		if (cui->logarithmic) {
			val = log(val);
		}
		if (val != cui->adjustment->get_value()) {
			cui->adjustment->set_value (val);
		}
	}
	cui->ignore_change--;
}

void
PluginUI::control_port_toggled (ControlUI* cui)
{
	if (!cui->ignore_change) {
		insert.set_control_port (cui->port_index, cui->button->get_active());
	}
}

void
PluginUIWindow::plugin_going_away (ARDOUR::Redirect* ignored)
{
	_pluginui.stop_updating(0);
	delete_when_idle (this);
}

void
PluginUI::setting_selected()
{
	Entry* entry = combo.get_entry();
	if (entry->get_text().length() > 0) {
		if (!plugin.set_state(entry->get_text())) {
			// Just in case
			ARDOUR_UI::instance()->allow_focus (false);
		}
	}
}

void
PluginUI::save_plugin_setting ()
{
	ArdourPrompter prompter (true);
	prompter.set_prompt(_("Name for plugin settings:"));

	prompter.show_all();
	prompter.done.connect(Main::quit.slot());

	Main::run();

	if (prompter.status == Gtkmmext::Prompter::entered) {
		string name;

		prompter.get_result(name);

		if (name.length()) {
			if(plugin.save_preset(name)){
				combo.set_popdown_strings(plugin.get_ladspa_settings());
			}
		}
	}
}

gint
PluginUI::entry_focus_event (GdkEventFocus* ev)
{
	if (ev->in) {
		ARDOUR_UI::instance()->allow_focus (true);
	} else {
		ARDOUR_UI::instance()->allow_focus (false);
	}
	return TRUE;
}

void
PluginUI::bypass_toggled ()
{
	bool x;

	if ((x = bypass_button.get_active()) == insert.active()) {
		insert.set_active (!x, this);
	}
}

void
PluginUI::redirect_active_changed (Redirect* r, void* src)
{
	bypass_button.set_active (!r->active());
}

gint
PluginUI::start_updating (GdkEventAny *ev)
{
	if (output_controls.size() > 0 ) {
		screen_update_connection.disconnect();
		screen_update_connection = ARDOUR_UI::instance()->RapidScreenUpdate.connect 
			(slot (*this, &PluginUI::output_update));
	}
	return FALSE;
}

gint 
PluginUI::stop_updating (GdkEventAny *ev)
{
	if (output_controls.size() > 0 ) {
		screen_update_connection.disconnect();
	}
	return FALSE;
}

void
PluginUI::output_update ()
{
	for (vector<ControlUI*>::iterator i = output_controls.begin(); i != output_controls.end(); ++i) {
		LADSPA_Data val = plugin.get_control_port ((*i)->port_index);
		char buf[32];
		snprintf (buf, sizeof(buf), "%.2f", val);
		(*i)->display_label->set_text (buf);

		/* autoscaling for the meter */
		if ((*i)->meterinfo && (*i)->meterinfo->packed) {
			
			if ((*i)->meterinfo->min_unbound && val < (*i)->meterinfo->min) {
				(*i)->meterinfo->min = val;
			}
			
			if ((*i)->meterinfo->max_unbound && val > (*i)->meterinfo->max) {
				(*i)->meterinfo->max = val;
			} 
			
			if ((*i)->meterinfo->max > (*i)->meterinfo->min ) {
				LADSPA_Data lval = (val - (*i)->meterinfo->min) / ((*i)->meterinfo->max - (*i)->meterinfo->min) ;
				(*i)->meterinfo->meter->set (lval );
			}
		}
	}
}
